from typing import Iterable, Dict, TYPE_CHECKING

import numpy as np

from docarray import DocumentArray
from docarray.array.storage.base.getsetdel import BaseGetSetDelMixin
from docarray.array.storage.base.helper import Offset2ID
from docarray.array.storage.milvus.backend import (
    _always_true_expr,
    _ids_to_milvus_expr,
    _batch_list,
)

if TYPE_CHECKING:
    from docarray import Document, DocumentArray


class GetSetDelMixin(BaseGetSetDelMixin):
    def _get_doc_by_id(self, _id: str) -> 'Document':
        # to be implemented
        return self._get_docs_by_ids([_id])[0]

    def _del_doc_by_id(self, _id: str):
        # to be implemented
        self._del_docs_by_ids([_id])

    def _set_doc_by_id(self, _id: str, value: 'Document', **kwargs):
        # to be implemented
        self._set_docs_by_ids([_id], [value], None, **kwargs)

    def _load_offset2ids(self):
        if self._list_like:
            collection = self._offset2id_collection
            kwargs = self._update_kwargs_from_config('consistency_level', **dict())
            with self.loaded_collection(collection):
                res = collection.query(
                    expr=_always_true_expr('document_id'),
                    output_fields=['offset', 'document_id'],
                    **kwargs,
                )
            sorted_res = sorted(res, key=lambda k: int(k['offset']))
            self._offset2ids = Offset2ID([r['document_id'] for r in sorted_res])
        else:
            self._offset2ids = Offset2ID([], list_like=self._list_like)

    def _save_offset2ids(self):
        if self._list_like:
            # delete old entries
            self._clear_offset2ids_milvus()
            # insert current entries
            ids = self._offset2ids.ids
            if not ids:
                return
            offsets = [str(i) for i in range(len(ids))]
            dummy_vectors = [np.zeros(1) for _ in range(len(ids))]
            collection = self._offset2id_collection
            collection.insert([offsets, ids, dummy_vectors])

    def _get_docs_by_ids(self, ids: 'Iterable[str]', **kwargs) -> 'DocumentArray':
        if not ids:
            return DocumentArray()
        ids = list(ids)
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        kwargs = self._update_kwargs_from_config('batch_size', **kwargs)
        with self.loaded_collection():
            docs = DocumentArray()
            for id_batch in _batch_list(ids, kwargs['batch_size']):
                res = self._collection.query(
                    expr=f'document_id in {_ids_to_milvus_expr(id_batch)}',
                    output_fields=['serialized'],
                    **kwargs,
                )
            if not res:
                raise KeyError(f'No documents found for ids {ids}')
            docs.extend(self._docs_from_query_response(res))
        # sort output docs according to input id sorting
        return DocumentArray([docs[d] for d in ids])

    def _del_docs_by_ids(self, ids: 'Iterable[str]', **kwargs) -> 'DocumentArray':
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        kwargs = self._update_kwargs_from_config('batch_size', **kwargs)
        for id_batch in _batch_list(list(ids), kwargs['batch_size']):
            self._collection.delete(
                expr=f'document_id in {_ids_to_milvus_expr(id_batch)}', **kwargs
            )

    def _set_docs_by_ids(
        self, ids, docs: 'Iterable[Document]', mismatch_ids: 'Dict', **kwargs
    ):
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        kwargs = self._update_kwargs_from_config('batch_size', **kwargs)
        # delete old entries
        for id_batch in _batch_list(list(ids), kwargs['batch_size']):
            self._collection.delete(
                expr=f'document_id in {_ids_to_milvus_expr(id_batch)}',
                **kwargs,
            )
        for docs_batch in _batch_list(list(docs), kwargs['batch_size']):
            # insert new entries
            payload = self._docs_to_milvus_payload(docs_batch)
            self._collection.insert(payload, **kwargs)

    def _clear_storage(self):
        self._collection.drop()
        self._create_or_reuse_collection()
        self._clear_offset2ids_milvus()

    def _clear_offset2ids_milvus(self):
        self._offset2id_collection.drop()
        self._create_or_reuse_offset2id_collection()



import base64
import io
import math
import struct
import warnings
from typing import Optional, Tuple, Union, BinaryIO, TYPE_CHECKING, List

import numpy as np

from docarray.document.mixins.helper import _get_file_context, _uri_to_blob

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import T
    from PIL.Image import Image as PILImage


class ImageDataMixin:
    """Provide helper functions for :class:`Document` to support image data."""

    def set_image_tensor_channel_axis(
        self: 'T', original_channel_axis: int, new_channel_axis: int
    ) -> 'T':
        """Move the channel axis of the image :attr:`.tensor` inplace.

        :param original_channel_axis: the original axis of the channel
        :param new_channel_axis: the new axis of the channel

        :return: itself after processed
        """
        self.tensor = _move_channel_axis(
            self.tensor, original_channel_axis, new_channel_axis
        )
        return self

    def load_pil_image_to_datauri(self, image: 'PILImage'):
        """Convert a pillow image into a datauri with header `data:image/png`.

        :param image: a pillow image
        :return: itself after processed
        """
        buffer = io.BytesIO()
        image.save(buffer, format='png')
        self.uri = (
            f'data:image/png;base64,' + base64.b64encode(buffer.getvalue()).decode()
        )
        return self

    def convert_blob_to_image_tensor(
        self: 'T',
        width: Optional[int] = None,
        height: Optional[int] = None,
        channel_axis: int = -1,
    ) -> 'T':
        """Convert an image :attr:`.blob` to a ndarray :attr:`.tensor`.

        :param width: the width of the image tensor.
        :param height: the height of the tensor.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """
        tensor = _to_image_tensor(io.BytesIO(self.blob), width=width, height=height)
        tensor = _move_channel_axis(tensor, original_channel_axis=channel_axis)
        self.tensor = tensor
        return self

    def convert_image_tensor_to_uri(
        self: 'T', channel_axis: int = -1, image_format: str = 'png'
    ) -> 'T':
        """Assuming :attr:`.tensor` is a _valid_ image, set :attr:`uri` accordingly

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param image_format: either `png` or `jpeg`
        :return: itself after processed
        """

        tensor = _move_channel_axis(self.tensor, original_channel_axis=channel_axis)
        _bytes = _to_image_buffer(tensor, image_format)
        self.uri = (
            f'data:image/{image_format};base64,' + base64.b64encode(_bytes).decode()
        )
        return self

    def convert_image_tensor_to_blob(
        self: 'T', channel_axis: int = -1, image_format: str = 'png'
    ) -> 'T':
        """Assuming :attr:`.tensor` is a _valid_ image, set :attr:`blob` accordingly

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param image_format: either `png` or `jpeg`
        :return: itself after processed
        """
        tensor = _move_channel_axis(self.tensor, original_channel_axis=channel_axis)
        self.blob = _to_image_buffer(tensor, image_format)
        return self

    def set_image_tensor_resample(
        self: 'T',
        ratio: float,
        channel_axis: int = -1,
    ) -> 'T':
        """
        Resample the image :attr:`.tensor` into different size inplace.

        :param ratio: scale ratio of the resampled image tensor.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :return: itself after processed
        """
        tensor = _move_channel_axis(self.tensor, channel_axis, -1)
        in_rows, in_cols, n_in = tensor.shape
        in_rows, in_cols = int(in_rows * ratio), int(in_cols * ratio)
        self.set_image_tensor_shape((in_rows, in_cols), channel_axis)

        return self

    def set_image_tensor_shape(
        self: 'T',
        shape: Tuple[int, int],
        channel_axis: int = -1,
    ) -> 'T':
        """Resample the image :attr:`.tensor` into different size inplace.

        If your current image tensor has shape ``[H,W,C]``, then the new tensor will be ``[*shape, C]``

        :param shape: the new shape of the image tensor.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """
        tensor = _move_channel_axis(self.tensor, channel_axis, -1)
        out_rows, out_cols = shape
        in_rows, in_cols, n_in = tensor.shape

        # compute coordinates to resample
        x = np.tile(np.linspace(0, in_cols - 2, out_cols), out_rows)
        y = np.repeat(np.linspace(0, in_rows - 2, out_rows), out_cols)

        # resample each image
        r = _nn_interpolate_2D(tensor, x, y)
        tensor = r.reshape(out_rows, out_cols, n_in)
        self.tensor = _move_channel_axis(tensor, -1, channel_axis)

        return self

    def save_image_tensor_to_file(
        self: 'T',
        file: Union[str, BinaryIO],
        channel_axis: int = -1,
        image_format: str = 'png',
    ) -> 'T':
        """Save :attr:`.tensor` into a file

        :param file: File or filename to which the data is saved.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param image_format: either `png` or `jpeg`

        :return: itself after processed
        """

        if isinstance(file, str) and not file.endswith(image_format.lower()):
            warnings.warn(
                f'your output file extension `{file}` does not match your output image format `{image_format}`. '
                f'This may result unreadable image file. Please double check your file name or `image_format`.'
            )

        fp = _get_file_context(file)
        with fp:
            tensor = _move_channel_axis(self.tensor, channel_axis, -1)
            buffer = _to_image_buffer(tensor, image_format)
            fp.write(buffer)
        return self

    def load_uri_to_image_tensor(
        self: 'T',
        width: Optional[int] = None,
        height: Optional[int] = None,
        channel_axis: int = -1,
        **kwargs,
    ) -> 'T':
        """Convert the image-like :attr:`.uri` into :attr:`.tensor`

        :param width: the width of the image tensor.
        :param height: the height of the tensor.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param kwargs: keyword arguments to pass to `:meth:_uri_to_blob` such as timeout

        :return: itself after processed
        """
        from docarray import Document, DocumentArray

        buffer = _uri_to_blob(self.uri, **kwargs)
        tensor = _to_image_tensor(io.BytesIO(buffer), width=width, height=height)

        if isinstance(tensor, np.ndarray):
            self.tensor = _move_channel_axis(tensor, original_channel_axis=channel_axis)
        elif isinstance(tensor, List):
            self.chunks = DocumentArray(
                [
                    Document(
                        tensor=_move_channel_axis(
                            tensor[i], original_channel_axis=channel_axis
                        )
                    )
                    for i in range(len(tensor))
                ]
            )

        return self

    def set_image_tensor_inv_normalization(
        self: 'T',
        channel_axis: int = -1,
        img_mean: Tuple[float] = (0.485, 0.456, 0.406),
        img_std: Tuple[float] = (0.229, 0.224, 0.225),
    ) -> 'T':
        """Inverse the normalization of a float32 image :attr:`.tensor` into a uint8 image :attr:`.tensor` inplace.

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param img_mean: the mean of all images
        :param img_std: the standard deviation of all images
        :return: itself after processed
        """
        if self.tensor.dtype == np.float32 and self.tensor.ndim == 3:
            tensor = _move_channel_axis(self.tensor, channel_axis, 0)
            mean = np.asarray(img_mean, dtype=np.float32)
            std = np.asarray(img_std, dtype=np.float32)
            tensor = ((tensor * std[:, None, None] + mean[:, None, None]) * 255).astype(
                np.uint8
            )
            # set back channel to original
            tensor = _move_channel_axis(tensor, 0, channel_axis)
            self.tensor = tensor
        else:
            raise ValueError(
                f'`tensor` must be a float32 ndarray with ndim=3, but receiving {self.tensor.dtype} with ndim={self.tensor.ndim}'
            )
        return self

    def set_image_tensor_normalization(
        self: 'T',
        channel_axis: int = -1,
        img_mean: Tuple[float] = (0.485, 0.456, 0.406),
        img_std: Tuple[float] = (0.229, 0.224, 0.225),
    ) -> 'T':
        """Normalize a uint8 image :attr:`.tensor` into a float32 image :attr:`.tensor` inplace.

        Applies normalization to the color channels of the images.
        By default, the normalization uses mean = [0.485, 0.456, 0.406] and std = [0.229, 0.224, 0.225], which are standard values computed on millions of images. If you want to train from scratch on your own dataset, you can calculate the new
        mean and std. Otherwise, using the Imagenet pretrained model with its own mean and std is recommended.

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param img_mean: the means of all images: [mean_r, mean_g, mean_b]
        :param img_std: the standard deviations of all images: [std_r, std_g, std_b]
        :return: itself after processed

        .. warning::
            Please do NOT generalize this function to gray scale, black/white image, it does not make any sense for
            non RGB image. if you look at their MNIST examples, the mean and stddev are 1-dimensional
            (since the inputs are greyscale-- no RGB channels).


        """
        if self.tensor.dtype == np.uint8 and self.tensor.ndim == 3:
            tensor = (self.tensor / 255.0).astype(np.float32)
            tensor = _move_channel_axis(tensor, channel_axis, 0)
            mean = np.asarray(img_mean, dtype=np.float32)
            std = np.asarray(img_std, dtype=np.float32)
            tensor = (tensor - mean[:, None, None]) / std[:, None, None]
            # set back channel to original
            tensor = _move_channel_axis(tensor, 0, channel_axis)
            self.tensor = tensor
        else:
            raise ValueError(
                f'`tensor` must be a uint8 ndarray with ndim=3, but receiving {self.tensor.dtype} with ndim={self.tensor.ndim}'
            )
        return self

    def convert_image_tensor_to_sliding_windows(
        self: 'T',
        window_shape: Tuple[int, int] = (64, 64),
        strides: Optional[Tuple[int, int]] = None,
        padding: bool = False,
        channel_axis: int = -1,
        as_chunks: bool = False,
    ) -> 'T':
        """Convert :attr:`.tensor` into a sliding window view with the given window shape :attr:`.tensor` inplace.

        :param window_shape: desired output size. If size is a sequence like (h, w), the output size will be matched to
            this. If size is an int, the output will have the same height and width as the `target_size`.
        :param strides: the strides between two neighboring sliding windows. `strides` is a sequence like (h, w), in
            which denote the strides on the vertical and the horizontal axis. When not given, using `window_shape`
        :param padding: If False, only patches which are fully contained in the input image are included. If True,
            all patches whose starting point is inside the input are included, and areas outside the input default to
            zero. The `padding` argument has no effect on the size of each patch, it determines how many patches are
            extracted. Default is False.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis.
        :param as_chunks: If set, each sliding window will be stored in the chunk of the current Document
        :return: Document itself after processed
        """
        window_h, window_w = window_shape
        stride_h, stride_w = strides or window_shape
        tensor = _move_channel_axis(self.tensor, channel_axis, -1)
        if padding:
            h, w, c = tensor.shape
            if h % stride_h:
                ext_h = window_h - h % stride_h
            else:
                ext_h = 0
            if w % stride_w:
                ext_w = window_w - w % stride_w
            else:
                ext_w = 0
            tensor = np.pad(
                tensor,
                ((0, ext_h), (0, ext_w), (0, 0)),
                mode='constant',
                constant_values=0,
            )
        h, w, c = tensor.shape
        row_step = tensor.strides[0]
        col_step = tensor.strides[1]

        expanded_img = np.lib.stride_tricks.as_strided(
            tensor,
            shape=(
                1 + math.ceil((h - window_h) / stride_h),
                1 + math.ceil((w - window_w) / stride_w),
                window_h,
                window_w,
                c,
            ),
            strides=(row_step * stride_h, col_step * stride_w, row_step, col_step, 1),
            writeable=False,
        )
        cur_loc_h, cur_loc_w = 0, 0
        if self.location:
            cur_loc_h, cur_loc_w = self.location[:2]

        bbox_locations = [
            (h * stride_h + cur_loc_h, w * stride_w + cur_loc_w, window_h, window_w)
            for h in range(expanded_img.shape[0])
            for w in range(expanded_img.shape[1])
        ]
        expanded_img = expanded_img.reshape((-1, window_h, window_w, c))
        if as_chunks:
            from docarray.document import Document

            for location, _tensor in zip(bbox_locations, expanded_img):
                self.chunks.append(
                    Document(
                        tensor=_move_channel_axis(_tensor, -1, channel_axis),
                        location=location,
                    )
                )
        else:
            self.tensor = _move_channel_axis(expanded_img, -1, channel_axis)
        return self


def _move_channel_axis(
    tensor: np.ndarray, original_channel_axis: int = -1, target_channel_axis: int = -1
) -> np.ndarray:
    """This will always make the channel axis to the last of the :attr:`.tensor`

    #noqa: DAR101
    #noqa: DAR201
    """
    if original_channel_axis != target_channel_axis:
        tensor = np.moveaxis(tensor, original_channel_axis, target_channel_axis)
    return tensor


def _to_image_tensor(
    source,
    width: Optional[int] = None,
    height: Optional[int] = None,
) -> Union[np.ndarray, List[np.array]]:
    """
    Convert an image blob to tensor or List of image tensors, if multiple images are stored in file.

    :param source: binary blob or file path
    :param width: the width of the image tensor.
    :param height: the height of the tensor.
    :return: image tensor or list of image tensors
    """
    from PIL import Image

    raw_img = Image.open(source)

    if width or height:
        new_width = width or raw_img.width
        new_height = height or raw_img.height

    # support multi page tiff images
    try:
        n_frames = raw_img.n_frames
    except AttributeError:
        n_frames = 1

    if n_frames > 1:

        frames = []
        for i in range(raw_img.n_frames):
            raw_img.seek(i)
            try:
                img = raw_img.convert('RGB')
            except:
                img = raw_img

            if width or height:
                frames.append(np.array(img.resize((new_width, new_height))))
            else:
                frames.append(np.array(img))

        return frames

    else:
        if width or height:
            raw_img = raw_img.resize((new_width, new_height))
        try:
            return np.array(raw_img.convert('RGB'))
        except:
            return np.array(raw_img)


def _to_image_buffer(arr: 'np.ndarray', image_format: str) -> bytes:
    """
    Convert image-ndarray to buffer bytes.

    :param arr: Data representations of the png.
    :param image_format: `png` or `jpeg`
    :return: Png in buffer bytes.

    ..note::
        if both :attr:`width` and :attr:`height` were provided, will not resize. Otherwise, will get image size
        by :attr:`arr` shape and apply resize method :attr:`resize_method`.
    """

    if image_format not in ('png', 'jpeg', 'jpg'):
        raise ValueError(
            f'image_format must be either `png` or `jpeg`, receiving `{image_format}`'
        )
    if image_format == 'jpg':
        image_format = 'jpeg'  # unify it to ISO standard

    arr = arr.astype(np.uint8).squeeze()

    if arr.ndim == 1:
        # note this should be only used for MNIST/FashionMNIST dataset, because of the nature of these two datasets
        # no other image data should flattened into 1-dim array.
        image_bytes = _png_to_buffer_1d(arr, 28, 28)
    elif arr.ndim == 2:
        from PIL import Image

        im = Image.fromarray(arr).convert('L')
        image_bytes = _pillow_image_to_buffer(im, image_format=image_format.upper())
    elif arr.ndim == 3:
        from PIL import Image

        im = Image.fromarray(arr).convert('RGB')
        image_bytes = _pillow_image_to_buffer(im, image_format=image_format.upper())
    else:
        raise ValueError(
            f'{arr.shape} ndarray can not be converted into an image buffer.'
        )

    return image_bytes


def _png_to_buffer_1d(arr: 'np.ndarray', width: int, height: int) -> bytes:
    import zlib

    pixels = []
    for p in arr[::-1]:
        pixels.extend([p, p, p, 255])
    buf = bytearray(pixels)

    # reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4
    raw_data = b''.join(
        b'\x00' + buf[span : span + width_byte_4]
        for span in range((height - 1) * width_byte_4, -1, -width_byte_4)
    )

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (
            struct.pack('!I', len(data))
            + chunk_head
            + struct.pack('!I', 0xFFFFFFFF & zlib.crc32(chunk_head))
        )

    png_bytes = b''.join(
        [
            b'\x89PNG\r\n\x1a\n',
            png_pack(b'IHDR', struct.pack('!2I5B', width, height, 8, 6, 0, 0, 0)),
            png_pack(b'IDAT', zlib.compress(raw_data, 9)),
            png_pack(b'IEND', b''),
        ]
    )

    return png_bytes


def _pillow_image_to_buffer(image, image_format: str) -> bytes:
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format=image_format)
    img_byte_arr = img_byte_arr.getvalue()
    return img_byte_arr


def _nn_interpolate_2D(X, x, y):
    nx, ny = np.around(x), np.around(y)
    nx = np.clip(nx, 0, X.shape[1] - 1).astype(int)
    ny = np.clip(ny, 0, X.shape[0] - 1).astype(int)
    return X[ny, nx, :]



from typing import Dict, Iterable, Sequence

from docarray import Document
from docarray.array.storage.base.getsetdel import BaseGetSetDelMixin
from docarray.array.storage.base.helper import Offset2ID


class GetSetDelMixin(BaseGetSetDelMixin):
    """Provide concrete implementation for ``__getitem__``, ``__setitem__``,
    and ``__delitem__`` for ``DocumentArrayRedis``"""

    def _get_doc_by_id(self, _id: str) -> 'Document':
        """Concrete implementation of base class' ``_get_doc_by_id``

        :param _id: the id of the document
        :return: the retrieved document from redis
        """
        try:
            result = self._client.hgetall(self._doc_prefix + _id)
            doc = Document.from_base64(result[b'blob'])
            return doc
        except Exception as ex:
            raise KeyError(_id) from ex

    def _get_docs_by_ids(self, ids: Sequence[str]) -> Iterable['Document']:
        """Concrete implementation of base class' ``_get_docs_by_ids``

        :param ids:  ids of the document
        :return: Iterable[Document]
        """

        accumulated_docs = []
        accumulated_docs_id_not_found = []

        if not ids:
            return accumulated_docs

        pipe = self._client.pipeline()
        for id in ids:
            pipe.hgetall(self._doc_prefix + id)

        results = pipe.execute()

        for i, result in enumerate(results):
            if result:
                accumulated_docs.append(Document.from_base64(result[b'blob']))
            else:
                accumulated_docs_id_not_found.append(ids[i])

        if accumulated_docs_id_not_found:
            raise KeyError(accumulated_docs_id_not_found, accumulated_docs)

        return accumulated_docs

    def _set_doc_by_id(self, _id: str, value: 'Document'):
        """Concrete implementation of base class' ``_set_doc_by_id``

        :param _id: the id of doc to update
        :param value: the document to update to
        """
        self._del_doc_by_id(_id)
        if _id != value.id:
            self._del_doc_by_id(value.id)

        payload = self._document_to_redis(value)
        self._client.hset(self._doc_prefix + value.id, mapping=payload)

    def _set_docs_by_ids(self, ids, docs: Iterable['Document'], mismatch_ids: Dict):
        """Overridden implementation of _set_docs_by_ids in order to add docs in batches and flush at the end

        :param ids: the ids used for indexing
        """
        for _id, doc in zip(ids, docs):
            self._del_doc_by_id(_id)
            if _id != doc.id:
                self._del_doc_by_id(doc.id)

        self._upload_batch(docs)

    def _del_doc_by_id(self, _id: str):
        """Concrete implementation of base class' ``_del_doc_by_id``

        :param _id: the id of the document to delete
        """
        if self._doc_id_exists(_id):
            self._client.delete(self._doc_prefix + _id)

    def _document_to_redis(self, doc: 'Document') -> Dict:
        extra_columns = {}

        for col, _ in self._config.columns.items():
            tag = doc.tags.get(col)
            if tag is not None:
                extra_columns[col] = int(tag) if isinstance(tag, bool) else tag

        if self._config.tag_indices:
            for index in self._config.tag_indices:
                text = doc.tags.get(index)
                if text is not None:
                    extra_columns[index] = text

        payload = {
            'id': doc.id,
            'embedding': self._map_embedding(doc.embedding),
            'blob': doc.to_base64(),
            **extra_columns,
        }

        if doc.text:
            payload['text'] = doc.text
        return payload

    def _load_offset2ids(self):
        if self._list_like:
            ids = self._get_offset2ids_meta()
            self._offset2ids = Offset2ID(ids, list_like=self._list_like)
        else:
            self._offset2ids = Offset2ID([], list_like=self._list_like)

    def _save_offset2ids(self):
        if self._list_like:
            self._update_offset2ids_meta()

    def _clear_storage(self):
        self._client.ft(index_name=self._config.index_name).dropindex(
            delete_documents=True
        )
        self._build_index(rebuild=True)
        self._client.delete(self._offset2id_key)



from typing import Iterable, Iterator, Union, TYPE_CHECKING
from docarray.array.storage.base.seqlike import BaseSequenceLikeMixin
from docarray.array.storage.milvus.backend import _batch_list
from docarray import Document


class SequenceLikeMixin(BaseSequenceLikeMixin):
    def __eq__(self, other):
        """Compare this object to the other, returns True if and only if other
        as the same type as self and other have the same Milvus Collections for data and offset2id

        :param other: the other object to check for equality
        :return: `True` if other is equal to self
        """
        return (
            type(self) is type(other)
            and self._collection.name == other._collection.name
            and self._offset2id_collection.name == other._offset2id_collection.name
            and self._config == other._config
        )

    def __contains__(self, x: Union[str, 'Document']):
        if isinstance(x, Document):
            x = x.id
        try:
            self._get_doc_by_id(x)
            return True
        except:
            return False

    def __repr__(self):
        return f'<DocumentArray[Milvus] (length={len(self)}) at {id(self)}>'

    def __add__(self, other: Union['Document', Iterable['Document']]):
        if isinstance(other, Document):
            self.append(other)
        else:
            self.extend(other)
        return self

    def insert(self, index: int, value: 'Document', **kwargs):
        self._set_doc_by_id(value.id, value, **kwargs)
        self._offset2ids.insert(index, value.id)

    def _append(self, value: 'Document', **kwargs):
        self._set_doc_by_id(value.id, value, **kwargs)
        self._offset2ids.append(value.id)

    def _extend(self, values: Iterable['Document'], **kwargs):
        docs = list(values)
        if not docs:
            return
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        kwargs = self._update_kwargs_from_config('batch_size', **kwargs)
        for docs_batch in _batch_list(list(docs), kwargs['batch_size']):
            payload = self._docs_to_milvus_payload(docs_batch)
            self._collection.insert(payload, **kwargs)
            self._offset2ids.extend([doc.id for doc in docs_batch])



import wave
from typing import Union, BinaryIO, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import T


class AudioDataMixin:
    """Provide helper functions for :class:`Document` to support audio data."""

    def save_audio_tensor_to_file(
        self: 'T',
        file: Union[str, BinaryIO],
        sample_rate: int = 44100,
        sample_width: int = 2,
    ) -> 'T':
        """Save :attr:`.tensor` into an wav file. Mono/stereo is preserved.

        :param file: if file is a string, open the file by that name, otherwise treat it as a file-like object.
        :param sample_rate: sampling frequency
        :param sample_width: sample width in bytes

        :return: Document itself after processed
        """
        # Convert to (little-endian) 16 bit integers.
        max_int16 = 2**15
        tensor = (self.tensor * max_int16).astype('<h')
        n_channels = 2 if self.tensor.ndim > 1 else 1

        with wave.open(file, 'w') as f:
            # 2 Channels.
            f.setnchannels(n_channels)
            # 2 bytes per sample.
            f.setsampwidth(sample_width)
            f.setframerate(sample_rate)
            f.writeframes(tensor.tobytes())
        return self

    def load_uri_to_audio_tensor(self: 'T') -> 'T':
        """Convert an audio :attr:`.uri` into :attr:`.tensor` inplace

        :return: Document itself after processed
        """
        with wave.open(
            self.uri
        ) as ifile:  #: note wave is Python built-in module https://docs.python.org/3/library/wave.html
            samples = ifile.getnframes()
            audio = ifile.readframes(samples)

            # Convert buffer to float32 using NumPy
            audio_as_np_int16 = np.frombuffer(audio, dtype=np.int16)
            audio_as_np_float32 = audio_as_np_int16.astype(np.float32)

            # Normalise float32 array so that values are between -1.0 and +1.0
            max_int16 = 2**15
            audio_normalised = audio_as_np_float32 / max_int16

            channels = ifile.getnchannels()
            if channels == 2:
                # 1 for mono, 2 for stereo
                audio_stereo = np.empty(
                    (int(len(audio_normalised) / channels), channels)
                )
                audio_stereo[:, 0] = audio_normalised[
                    range(0, len(audio_normalised), 2)
                ]
                audio_stereo[:, 1] = audio_normalised[
                    range(1, len(audio_normalised), 2)
                ]

                self.tensor = audio_stereo
            else:
                self.tensor = audio_normalised
            return self



# some implementations are adopted from https://github.com/ncoop57/cute_ranking/blob/main/cute_ranking/core.py
# the original code is licensed under Apache-2.0

from typing import List, Optional

import numpy as np


def _check_k(k):
    if k is not None and k < 1:
        raise ValueError(f'`k` must be >=1 or `None`')


def r_precision(binary_relevance: List[int], **kwargs) -> float:
    """R-Precision determines the precision in the fist R documents, where R is the
    number of documents relevant to the query.

    Relevance is considered binary by this function (nonzero is relevant).

    Please note, that it is necessary to provide relevance scores for all documents,
    i.e., the calculated metric is wrong, if you apply it on the Top-K scores only.

    .. seealso::
        https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#R-precision

    :param binary_relevance: binary relevancy in rank order
    :return: R-Precision
    """
    binary_relevance = np.array(binary_relevance) != 0
    z = binary_relevance.nonzero()[0]
    if not z.size:
        return 0.0
    return float(np.mean(binary_relevance[: z.size]))


def precision_at_k(
    binary_relevance: List[int], k: Optional[int] = None, **kwargs
) -> float:
    """Precision @K.
    If `binary_relevance` is empty, 0.0 is returned.

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :return: precision @k
    """
    _check_k(k)
    if len(binary_relevance) == 0:
        return 0.0
    binary_relevance = np.array(binary_relevance)[:k] != 0
    return float(np.mean(binary_relevance))


def hit_at_k(binary_relevance: List[int], k: Optional[int] = None, **kwargs) -> int:
    """Score is percentage of first relevant item in list that occur

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :return: hit @k if hit return 1 else 0
    """
    _check_k(k)
    return 1 if np.sum(binary_relevance[:k]) > 0 else 0


def average_precision(binary_relevance: List[int], **kwargs) -> float:
    """Score is average precision (area under PR curve)
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :return: Average precision
    """
    r = np.array(binary_relevance) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.0
    return float(np.mean(out))


def reciprocal_rank(binary_relevance: List[int], **kwargs) -> float:
    """Score is reciprocal of the rank of the first relevant item

    :param binary_relevance: binary relevancy in rank order
    :return: Average precision
    """
    rs = np.array(binary_relevance).nonzero()[0]
    return 1.0 / (rs[0] + 1) if rs.size else 0.0


def recall_at_k(
    binary_relevance: List[int], max_rel: int, k: Optional[int] = None, **kwargs
) -> float:
    """Score is recall after all relevant documents have been retrieved
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :param max_rel: Maximum number of documents that can be relevant
    :return: Recall score
    """
    _check_k(k)
    binary_relevance = np.array(binary_relevance[:k]) != 0
    if max_rel is None:
        raise ValueError('The metric recall_at_k requires a max_rel parameter')
    if np.sum(binary_relevance) > max_rel:
        raise ValueError(f'Number of relevant Documents retrieved > {max_rel}')
    return np.sum(binary_relevance) / max_rel


def f1_score_at_k(
    binary_relevance: List[int], max_rel: int, k: Optional[int] = None, **kwargs
) -> float:
    """Score is harmonic mean of precision and recall
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :param max_rel: Maximum number of documents that can be relevant
    :return: F1 score @ k
    """
    _check_k(k)
    p = precision_at_k(binary_relevance, k)
    r = recall_at_k(binary_relevance, max_rel, k)
    if (p + r) > 0:
        return 2 * p * r / (p + r)
    else:
        return 0.0


def dcg_at_k(
    relevance: List[float], method: int = 0, k: Optional[int] = None, **kwargs
):
    """Score is discounted cumulative gain (dcg)
    Relevance is positive real values. Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf

    :param relevance: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    :param k: measured on top-k
    :param method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]
    :return: Discounted cumulative gain
    """
    _check_k(k)
    r = np.asfarray(relevance)[:k]
    if r.size:
        if method == 0:
            return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
        elif method == 1:
            return np.sum(r / np.log2(np.arange(2, r.size + 2)))
        else:
            raise ValueError('method must be 0 or 1.')
    return 0.0


def ndcg_at_k(
    relevance: List[float], method: int = 0, k: Optional[int] = None, **kwargs
):
    """Calculates a normalized discounted cumulative gain (ndcg).
    Relevance values can be positive real values. However, one can also use binary
    scores as in other evaluation methods.

    Please note, that it is necessary to provide relevance scores for all documents,
    i.e., the calculated metric is wrong, if you apply it on the Top-K scores only.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf

    :param relevance: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    :param k: measured on top-k
    :param method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]
    :return: Normalized discounted cumulative gain
    """
    _check_k(k)
    dcg_max = dcg_at_k(sorted(relevance, reverse=True), method=method, k=k)
    if not dcg_max:
        return 0.0
    return dcg_at_k(relevance, method=method, k=k) / dcg_max



from enum import Enum


class DocumentMetadata(str, Enum):
    MULTI_MODAL_SCHEMA = 'multi_modal_schema'
    IMAGE_TYPE = 'image_type'
    IMAGE_URI = 'image_uri'
    AUDIO_TYPE = 'audio_type'
    VIDEO_TYPE = 'video_type'
    MESH_TYPE = 'mesh_type'
    JSON_TYPE = 'json_type'
    BLOB_TYPE = 'blob_type'


class AttributeType(str, Enum):
    DOCUMENT = 'document'
    PRIMITIVE = 'primitive'
    ITERABLE_PRIMITIVE = 'iterable_primitive'
    ITERABLE_DOCUMENT = 'iterable_document'
    NESTED = 'nested'
    ITERABLE_NESTED = 'iterable_nested'


class ImageType(str, Enum):
    PIL = 'PIL'
    URI = 'uri'
    NDARRAY = 'ndarray'



from .document import DocumentArray

from .storage.milvus import StorageMixins, MilvusConfig

__all__ = ['MilvusConfig', 'DocumentArrayMilvus']


class DocumentArrayMilvus(StorageMixins, DocumentArray):
    """
    DocumentArray that stores Documents in a `Milvus <https://milvus.io//>`_ vector search engine.

    .. note::
        This DocumentArray requires `pymilvus`. You can install it via `pip install "docarray[milvus]"`.

        To use Milvus as storage backend, a Milvus service needs to be running on your machine.

    With this implementation, :meth:`match` and :meth:`find` perform fast (approximate) vector search.
    Additionally, search with filters is supported.

    Example usage:

    .. code-block:: python

        from docarray import DocumentArray

        # connect to running Milvus service with default configuration (address: http://localhost:19530)
        da = DocumentArray(storage='milvus', config={'n_dim': 10})

        # connect to a previously persisted DocumentArrayMilvus by specifying collection_name, host, and port
        da = DocumentArray(
            storage='milvus',
            config={
                'collection_name': 'persisted',
                'host': 'localhost',
                'port': '19530',
                'n_dim': 10,
            },
        )


    .. seealso::
        For further details, see our :ref:`user guide <milvus>`.
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)



import copy
import uuid
from typing import Optional, TYPE_CHECKING, Union, Dict, Iterable, List, Tuple
from dataclasses import dataclass, field
import re

import numpy as np
from pymilvus import (
    connections,
    Collection,
    FieldSchema,
    DataType,
    CollectionSchema,
    has_collection,
    loading_progress,
)

from docarray import Document, DocumentArray
from docarray.array.storage.base.backend import BaseBackendMixin, TypeMap
from docarray.helper import dataclass_from_dict, _safe_cast_int
from docarray.score import NamedScore

if TYPE_CHECKING:
    from docarray.typing import (
        DocumentArraySourceType,
    )


ID_VARCHAR_LEN = 1024
SERIALIZED_VARCHAR_LEN = (
    65_535  # 65_535 is the maximum that Milvus allows for a VARCHAR field
)
COLUMN_VARCHAR_LEN = 1024
OFFSET_VARCHAR_LEN = 1024


def _always_true_expr(primary_key: str) -> str:
    """
    Returns a Milvus expression that is always true, thus allowing for the retrieval of all entries in a Collection
    Assumes that the primary key is of type DataType.VARCHAR

    :param primary_key: the name of the primary key
    :return: a Milvus expression that is always true for that primary key
    """
    return f'({primary_key} in ["1"]) or ({primary_key} not in ["1"])'


def _ids_to_milvus_expr(ids):
    ids = ['"' + _id + '"' for _id in ids]
    return '[' + ','.join(ids) + ']'


def _batch_list(l: List, batch_size: int):
    """Iterates over a list in batches of size batch_size"""
    if batch_size < 1:
        yield l
        return
    l_len = len(l)
    for ndx in range(0, l_len, batch_size):
        yield l[ndx : min(ndx + batch_size, l_len)]


def _sanitize_collection_name(name):
    """Removes all chars that are not allowed in a Milvus collection name.
    Thus, it removes all chars that are not alphanumeric or an underscore.

    :param name: the collection name to sanitize
    :return: the sanitized collection name.
    """
    return ''.join(
        re.findall('[a-zA-Z0-9_]', name)
    )  # remove everything that is not a letter, number or underscore


@dataclass
class MilvusConfig:
    n_dim: int
    collection_name: str = None
    host: str = 'localhost'
    port: Optional[Union[str, int]] = 19530  # 19530 for gRPC, 9091 for HTTP
    distance: str = 'IP'  # metric_type in milvus
    index_type: str = 'HNSW'
    index_params: Dict = field(
        default_factory=lambda: {
            'M': 4,
            'efConstruction': 200,
        }
    )  # passed to milvus at index creation time. The default assumes 'HNSW' index type
    collection_config: Dict = field(
        default_factory=dict
    )  # passed to milvus at collection creation time
    serialize_config: Dict = field(default_factory=dict)
    consistency_level: str = 'Session'
    batch_size: int = -1
    columns: Optional[Union[List[Tuple[str, str]], Dict[str, str]]] = None
    list_like: bool = True
    root_id: bool = True


class BackendMixin(BaseBackendMixin):

    TYPE_MAP = {
        'str': TypeMap(type=DataType.VARCHAR, converter=str),
        'float': TypeMap(
            type=DataType.DOUBLE, converter=float
        ),  # it doesn't like DataType.FLOAT type, perhaps because python floats are double precision?
        'double': TypeMap(type=DataType.DOUBLE, converter=float),
        'int': TypeMap(type=DataType.INT64, converter=_safe_cast_int),
        'bool': TypeMap(type=DataType.BOOL, converter=bool),
    }

    def _init_storage(
        self,
        _docs: Optional['DocumentArraySourceType'] = None,
        config: Optional[Union[MilvusConfig, Dict]] = None,
        **kwargs,
    ):
        config = copy.deepcopy(config)
        if not config:
            raise ValueError('Empty config is not allowed for Milvus storage')
        elif isinstance(config, dict):
            config = dataclass_from_dict(MilvusConfig, config)

        if config.collection_name is None:
            id = uuid.uuid4().hex
            config.collection_name = 'docarray__' + id
        self._list_like = config.list_like
        self._config = config
        self._config.columns = self._normalize_columns(self._config.columns)

        self._connection_alias = f'docarray_{config.host}_{config.port}'
        connections.connect(
            alias=self._connection_alias, host=config.host, port=config.port
        )

        self._collection = self._create_or_reuse_collection()
        self._offset2id_collection = self._create_or_reuse_offset2id_collection()
        self._build_index()
        super()._init_storage(**kwargs)

        # To align with Sqlite behavior; if `docs` is not `None` and table name
        # is provided, :class:`DocumentArraySqlite` will clear the existing
        # table and load the given `docs`
        if _docs is None:
            return

        self.clear()
        if isinstance(_docs, Iterable):
            self.extend(_docs)
        else:
            if isinstance(_docs, Document):
                self.append(_docs)

    def _create_or_reuse_collection(self):
        if has_collection(self._config.collection_name, using=self._connection_alias):
            return Collection(
                self._config.collection_name, using=self._connection_alias
            )

        document_id = FieldSchema(
            name='document_id',
            dtype=DataType.VARCHAR,
            max_length=ID_VARCHAR_LEN,
            is_primary=True,
        )
        embedding = FieldSchema(
            name='embedding', dtype=DataType.FLOAT_VECTOR, dim=self._config.n_dim
        )
        serialized = FieldSchema(
            name='serialized', dtype=DataType.VARCHAR, max_length=SERIALIZED_VARCHAR_LEN
        )

        additional_columns = []
        for col, coltype in self._config.columns.items():
            mapped_type = self._map_type(coltype)
            if mapped_type == DataType.VARCHAR:
                field_ = FieldSchema(
                    name=col, dtype=mapped_type, max_length=COLUMN_VARCHAR_LEN
                )
            else:
                field_ = FieldSchema(name=col, dtype=mapped_type)
            additional_columns.append(field_)

        schema = CollectionSchema(
            fields=[document_id, embedding, serialized, *additional_columns],
            description='DocumentArray collection schema',
        )
        return Collection(
            name=self._config.collection_name,
            schema=schema,
            using=self._connection_alias,
            **self._config.collection_config,
        )

    def _build_index(self):
        index_params = {
            'metric_type': self._config.distance,
            'index_type': self._config.index_type,
            'params': self._config.index_params,
        }
        self._collection.create_index(field_name='embedding', index_params=index_params)

    def _create_or_reuse_offset2id_collection(self):
        if has_collection(
            self._config.collection_name + '_offset2id', using=self._connection_alias
        ):
            return Collection(
                self._config.collection_name + '_offset2id',
                using=self._connection_alias,
            )

        document_id = FieldSchema(
            name='document_id', dtype=DataType.VARCHAR, max_length=ID_VARCHAR_LEN
        )
        offset = FieldSchema(
            name='offset',
            dtype=DataType.VARCHAR,
            max_length=OFFSET_VARCHAR_LEN,
            is_primary=True,
        )
        dummy_vector = FieldSchema(
            name='dummy_vector', dtype=DataType.FLOAT_VECTOR, dim=1
        )

        schema = CollectionSchema(
            fields=[offset, document_id, dummy_vector],
            description='offset2id for DocumentArray',
        )

        return Collection(
            name=self._config.collection_name + '_offset2id',
            schema=schema,
            using=self._connection_alias,
            # **self._config.collection_config,  # we probably don't want to apply the same config here
        )

    def _ensure_unique_config(
        self,
        config_root: dict,
        config_subindex: dict,
        config_joined: dict,
        subindex_name: str,
    ) -> dict:
        if 'collection_name' not in config_subindex:
            config_joined['collection_name'] = _sanitize_collection_name(
                config_joined['collection_name'] + '_subindex_' + subindex_name
            )
        return config_joined

    def _doc_to_milvus_payload(self, doc):
        return self._docs_to_milvus_payload([doc])

    def _docs_to_milvus_payload(self, docs: 'Iterable[Document]'):
        extra_columns = [
            [self._map_column(doc.tags.get(col), col_type) for doc in docs]
            for col, col_type in self._config.columns.items()
        ]
        return [
            [doc.id for doc in docs],
            [self._map_embedding(doc.embedding) for doc in docs],
            [doc.to_base64(**self._config.serialize_config) for doc in docs],
            *extra_columns,
        ]

    @staticmethod
    def _docs_from_query_response(response):
        return DocumentArray([Document.from_base64(d['serialized']) for d in response])

    @staticmethod
    def _docs_from_search_response(responses, distance: str) -> 'List[DocumentArray]':
        das = []
        for r in responses:
            da = []
            for hit in r:
                doc = Document.from_base64(hit.entity.get('serialized'))
                doc.scores[distance] = NamedScore(value=hit.score)
                da.append(doc)
            das.append(DocumentArray(da))
        return das

    def _update_kwargs_from_config(self, field_to_update, **kwargs):
        kwargs_field_value = kwargs.get(field_to_update, None)
        config_field_value = getattr(self._config, field_to_update, None)

        if (
            kwargs_field_value is not None or config_field_value is None
        ):  # no need to update
            return kwargs

        kwargs[field_to_update] = config_field_value
        return kwargs

    def _map_embedding(self, embedding):
        if embedding is not None:
            from docarray.math.ndarray import to_numpy_array

            embedding = to_numpy_array(embedding)

            if embedding.ndim > 1:
                embedding = np.asarray(embedding).squeeze()
        else:
            embedding = np.zeros(self._config.n_dim)
        return embedding

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['_collection']
        del d['_offset2id_collection']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        connections.connect(
            alias=self._connection_alias, host=self._config.host, port=self._config.port
        )
        self._collection = self._create_or_reuse_collection()
        self._offset2id_collection = self._create_or_reuse_offset2id_collection()

    def __enter__(self):
        _ = super().__enter__()
        self._collection.load()
        self._offset2id_collection.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._collection.release()
        self._offset2id_collection.release()
        super().__exit__(exc_type, exc_val, exc_tb)

    def loaded_collection(self, collection=None):
        """
        Context manager to load a collection and release it after the context is exited.
        If the collection is already loaded when entering, it will not be released while exiting.

        :param collection: the collection to load. If None, the main collection of this indexer is used.
        :return: Context manager for the provided collection.
        """

        class LoadedCollectionManager:
            def __init__(self, coll, connection_alias):
                self._collection = coll
                self._loaded_when_enter = False
                self._connection_alias = connection_alias

            def __enter__(self):
                self._loaded_when_enter = (
                    loading_progress(
                        self._collection.name, using=self._connection_alias
                    )['loading_progress']
                    != '0%'
                )
                self._collection.load()
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                if not self._loaded_when_enter:
                    self._collection.release()

        return LoadedCollectionManager(
            collection if collection else self._collection, self._connection_alias
        )



import copy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from docarray import Document
from docarray.array.storage.base.backend import BaseBackendMixin, TypeMap
from docarray.helper import dataclass_from_dict, random_identity, filter_dict

from redis import Redis
from redis.commands.search.field import NumericField, TextField, VectorField, GeoField
from redis.commands.search.indexDefinition import IndexDefinition

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import ArrayType, DocumentArraySourceType


@dataclass
class RedisConfig:
    n_dim: int
    host: str = field(default='localhost')
    port: int = field(default=6379)
    index_name: Optional[str] = None
    list_like: bool = True
    update_schema: bool = field(default=True)
    distance: str = field(default='COSINE')
    redis_config: Dict[str, Any] = field(default_factory=dict)
    index_text: bool = field(default=False)
    tag_indices: List[str] = field(default_factory=list)
    batch_size: int = field(default=64)
    method: str = field(default='HNSW')
    ef_construction: Optional[int] = None
    m: Optional[int] = None
    ef_runtime: Optional[int] = None
    block_size: Optional[int] = None
    initial_cap: Optional[int] = None
    columns: Optional[Union[List[Tuple[str, str]], Dict[str, str]]] = None
    root_id: bool = True


class BackendMixin(BaseBackendMixin):
    """Provide necessary functions to enable this storage backend."""

    TYPE_MAP = {
        'str': TypeMap(type='text', converter=TextField),
        'bytes': TypeMap(type='text', converter=TextField),
        'int': TypeMap(type='integer', converter=NumericField),
        'float': TypeMap(type='float', converter=NumericField),
        'double': TypeMap(type='double', converter=NumericField),
        'long': TypeMap(type='long', converter=NumericField),
        'geo': TypeMap(type='geo', converter=GeoField),
    }

    def _init_storage(
        self,
        _docs: Optional['DocumentArraySourceType'] = None,
        config: Optional[Union[RedisConfig, Dict]] = None,
        **kwargs,
    ):
        config = copy.deepcopy(config)
        if not config:
            raise ValueError('Empty config is not allowed for Redis storage')
        elif isinstance(config, dict):
            config = dataclass_from_dict(RedisConfig, config)

        if config.distance not in ['L2', 'IP', 'COSINE']:
            raise ValueError(
                f'Expecting distance metric one of COSINE, L2 OR IP, got {config.distance} instead'
            )
        if config.method not in ['HNSW', 'FLAT']:
            raise ValueError(
                f'Expecting search method one of HNSW OR FLAT, got {config.method} instead'
            )

        if config.redis_config.get('decode_responses'):
            config.redis_config['decode_responses'] = False

        if config.index_name is None:
            config.index_name = 'index_name__' + random_identity()

        self._offset2id_key = config.index_name + '__offset2id'
        self._config = config
        self.n_dim = self._config.n_dim
        self._list_like = config.list_like
        self._doc_prefix = config.index_name + ':'
        self._config.columns = self._normalize_columns(self._config.columns)

        self._client = self._build_client()
        self._build_index()

        super()._init_storage(**kwargs)

        if _docs is None:
            return
        elif isinstance(_docs, Iterable):
            self.extend(_docs)
        elif isinstance(_docs, Document):
            self.append(_docs)

    def _build_client(self):
        client = Redis(
            host=self._config.host,
            port=self._config.port,
            **self._config.redis_config,
        )
        return client

    def _build_index(self, rebuild: bool = False):
        if self._config.update_schema or rebuild:
            if self._config.index_name.encode() in self._client.execute_command(
                'FT._LIST'
            ):
                self._client.ft(index_name=self._config.index_name).dropindex()

            schema = self._build_schema_from_redis_config()
            idef = IndexDefinition(prefix=[self._doc_prefix])
            self._client.ft(index_name=self._config.index_name).create_index(
                schema, definition=idef
            )

    def _ensure_unique_config(
        self,
        config_root: dict,
        config_subindex: dict,
        config_joined: dict,
        subindex_name: str,
    ) -> dict:
        if 'index_name' not in config_subindex:
            config_joined['index_name'] = (
                config_joined['index_name'] + '_subindex_' + subindex_name
            )
        return config_joined

    def _build_schema_from_redis_config(self):
        index_param = {
            'TYPE': 'FLOAT32',
            'DIM': self.n_dim,
            'DISTANCE_METRIC': self._config.distance,
        }

        if self._config.method == 'HNSW':
            index_options = {
                'M': self._config.m,
                'EF_CONSTRUCTION': self._config.ef_construction,
                'EF_RUNTIME': self._config.ef_runtime,
            }
            index_param.update(filter_dict(index_options))

        if self._config.method == 'FLAT':
            index_options = {'BLOCK_SIZE': self._config.block_size}
            index_param.update(filter_dict(index_options))

        if self._config.initial_cap:
            index_param['INITIAL_CAP'] = self._config.initial_cap
        schema = [VectorField('embedding', self._config.method, index_param)]

        if self._config.index_text:
            schema.append(TextField('text'))

        for index in self._config.tag_indices:
            schema.append(TextField(index))

        for col, coltype in self._config.columns.items():
            schema.append(self._map_column(col, coltype))

        return schema

    def _doc_id_exists(self, doc_id):
        return self._client.exists(self._doc_prefix + doc_id)

    def _map_embedding(self, embedding: 'ArrayType') -> bytes:
        if embedding is not None:
            from docarray.math.ndarray import to_numpy_array

            embedding = to_numpy_array(embedding)

            if embedding.ndim > 1:
                embedding = np.asarray(embedding).squeeze()
        else:
            embedding = np.zeros(self.n_dim)
        return embedding.astype(np.float32).tobytes()

    def _get_offset2ids_meta(self) -> List[str]:
        if not self._client.exists(self._offset2id_key):
            return []
        ids = self._client.lrange(self._offset2id_key, 0, -1)
        return [id.decode() for id in ids]

    def _update_offset2ids_meta(self):
        """Update the offset2ids in redis"""
        if self._client.exists(self._offset2id_key):
            self._client.delete(self._offset2id_key)
        if len(self._offset2ids.ids) > 0:
            self._client.rpush(self._offset2id_key, *self._offset2ids.ids)

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['_client']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._client = self._build_client()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: docarray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0e\x64ocarray.proto\x12\x08\x64ocarray\x1a\x1cgoogle/protobuf/struct.proto\"A\n\x11\x44\x65nseNdArrayProto\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\r\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\xb6\x01\n\x0cNdArrayProto\x12,\n\x05\x64\x65nse\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProtoH\x00\x12.\n\x06sparse\x18\x02 \x01(\x0b\x32\x1c.docarray.SparseNdArrayProtoH\x00\x12\x10\n\x08\x63ls_name\x18\x03 \x01(\t\x12+\n\nparameters\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructB\t\n\x07\x63ontent\"~\n\x12SparseNdArrayProto\x12,\n\x07indices\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12+\n\x06values\x18\x02 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12\r\n\x05shape\x18\x03 \x03(\r\"V\n\x0fNamedScoreProto\x12\r\n\x05value\x18\x01 \x01(\x02\x12\x0f\n\x07op_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06ref_id\x18\x04 \x01(\t\"\xed\x05\n\rDocumentProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x04\x62lob\x18\x02 \x01(\x0cH\x00\x12(\n\x06tensor\x18\x03 \x01(\x0b\x32\x16.docarray.NdArrayProtoH\x00\x12\x0e\n\x04text\x18\x04 \x01(\tH\x00\x12\x13\n\x0bgranularity\x18\x05 \x01(\r\x12\x11\n\tadjacency\x18\x06 \x01(\r\x12\x11\n\tparent_id\x18\x07 \x01(\t\x12\x0e\n\x06weight\x18\x08 \x01(\x02\x12\x0b\n\x03uri\x18\t \x01(\t\x12\x10\n\x08modality\x18\n \x01(\t\x12\x11\n\tmime_type\x18\x0b \x01(\t\x12\x0e\n\x06offset\x18\x0c \x01(\x02\x12\x10\n\x08location\x18\r \x03(\x02\x12\'\n\x06\x63hunks\x18\x0e \x03(\x0b\x32\x17.docarray.DocumentProto\x12(\n\x07matches\x18\x0f \x03(\x0b\x32\x17.docarray.DocumentProto\x12)\n\tembedding\x18\x10 \x01(\x0b\x32\x16.docarray.NdArrayProto\x12%\n\x04tags\x18\x11 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x33\n\x06scores\x18\x12 \x03(\x0b\x32#.docarray.DocumentProto.ScoresEntry\x12=\n\x0b\x65valuations\x18\x13 \x03(\x0b\x32(.docarray.DocumentProto.EvaluationsEntry\x12*\n\t_metadata\x18\x14 \x01(\x0b\x32\x17.google.protobuf.Struct\x1aH\n\x0bScoresEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x1aM\n\x10\x45valuationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x42\t\n\x07\x63ontent\";\n\x12\x44ocumentArrayProto\x12%\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x17.docarray.DocumentProtob\x06proto3'
)


_DENSENDARRAYPROTO = DESCRIPTOR.message_types_by_name['DenseNdArrayProto']
_NDARRAYPROTO = DESCRIPTOR.message_types_by_name['NdArrayProto']
_SPARSENDARRAYPROTO = DESCRIPTOR.message_types_by_name['SparseNdArrayProto']
_NAMEDSCOREPROTO = DESCRIPTOR.message_types_by_name['NamedScoreProto']
_DOCUMENTPROTO = DESCRIPTOR.message_types_by_name['DocumentProto']
_DOCUMENTPROTO_SCORESENTRY = _DOCUMENTPROTO.nested_types_by_name['ScoresEntry']
_DOCUMENTPROTO_EVALUATIONSENTRY = _DOCUMENTPROTO.nested_types_by_name[
    'EvaluationsEntry'
]
_DOCUMENTARRAYPROTO = DESCRIPTOR.message_types_by_name['DocumentArrayProto']
DenseNdArrayProto = _reflection.GeneratedProtocolMessageType(
    'DenseNdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _DENSENDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DenseNdArrayProto)
    },
)
_sym_db.RegisterMessage(DenseNdArrayProto)

NdArrayProto = _reflection.GeneratedProtocolMessageType(
    'NdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _NDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.NdArrayProto)
    },
)
_sym_db.RegisterMessage(NdArrayProto)

SparseNdArrayProto = _reflection.GeneratedProtocolMessageType(
    'SparseNdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _SPARSENDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.SparseNdArrayProto)
    },
)
_sym_db.RegisterMessage(SparseNdArrayProto)

NamedScoreProto = _reflection.GeneratedProtocolMessageType(
    'NamedScoreProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _NAMEDSCOREPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.NamedScoreProto)
    },
)
_sym_db.RegisterMessage(NamedScoreProto)

DocumentProto = _reflection.GeneratedProtocolMessageType(
    'DocumentProto',
    (_message.Message,),
    {
        'ScoresEntry': _reflection.GeneratedProtocolMessageType(
            'ScoresEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _DOCUMENTPROTO_SCORESENTRY,
                '__module__': 'docarray_pb2'
                # @@protoc_insertion_point(class_scope:docarray.DocumentProto.ScoresEntry)
            },
        ),
        'EvaluationsEntry': _reflection.GeneratedProtocolMessageType(
            'EvaluationsEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _DOCUMENTPROTO_EVALUATIONSENTRY,
                '__module__': 'docarray_pb2'
                # @@protoc_insertion_point(class_scope:docarray.DocumentProto.EvaluationsEntry)
            },
        ),
        'DESCRIPTOR': _DOCUMENTPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DocumentProto)
    },
)
_sym_db.RegisterMessage(DocumentProto)
_sym_db.RegisterMessage(DocumentProto.ScoresEntry)
_sym_db.RegisterMessage(DocumentProto.EvaluationsEntry)

DocumentArrayProto = _reflection.GeneratedProtocolMessageType(
    'DocumentArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _DOCUMENTARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DocumentArrayProto)
    },
)
_sym_db.RegisterMessage(DocumentArrayProto)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DOCUMENTPROTO_SCORESENTRY._options = None
    _DOCUMENTPROTO_SCORESENTRY._serialized_options = b'8\001'
    _DOCUMENTPROTO_EVALUATIONSENTRY._options = None
    _DOCUMENTPROTO_EVALUATIONSENTRY._serialized_options = b'8\001'
    _DENSENDARRAYPROTO._serialized_start = 58
    _DENSENDARRAYPROTO._serialized_end = 123
    _NDARRAYPROTO._serialized_start = 126
    _NDARRAYPROTO._serialized_end = 308
    _SPARSENDARRAYPROTO._serialized_start = 310
    _SPARSENDARRAYPROTO._serialized_end = 436
    _NAMEDSCOREPROTO._serialized_start = 438
    _NAMEDSCOREPROTO._serialized_end = 524
    _DOCUMENTPROTO._serialized_start = 527
    _DOCUMENTPROTO._serialized_end = 1276
    _DOCUMENTPROTO_SCORESENTRY._serialized_start = 1114
    _DOCUMENTPROTO_SCORESENTRY._serialized_end = 1186
    _DOCUMENTPROTO_EVALUATIONSENTRY._serialized_start = 1188
    _DOCUMENTPROTO_EVALUATIONSENTRY._serialized_end = 1265
    _DOCUMENTARRAYPROTO._serialized_start = 1278
    _DOCUMENTARRAYPROTO._serialized_end = 1337
# @@protoc_insertion_point(module_scope)



from typing import TYPE_CHECKING, TypeVar, List, Union, Optional, Dict, Sequence

if TYPE_CHECKING:
    import numpy as np
    import tensorflow
    import torch

    # Define the expected input type that your ANN search supports
    MilvusArrayType = TypeVar(
        'MilvusArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )
    from docarray import Document, DocumentArray


class FindMixin:
    def _find(
        self,
        query: 'MilvusArrayType',
        limit: int = 10,
        filter: Optional[Dict] = None,
        param=None,
        **kwargs
    ) -> List['DocumentArray']:
        """Returns `limit` approximate nearest neighbors given a batch of input queries.
        If the query is a single query, should return a DocumentArray, otherwise a list of DocumentArrays containing
        the closest Documents for each query.
        """
        if param is None:
            param = dict()
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        with self.loaded_collection():
            results = self._collection.search(
                data=query,
                anns_field='embedding',
                limit=limit,
                expr=filter,
                param=param,
                output_fields=['serialized'],
                **kwargs,
            )
        return self._docs_from_search_response(results, distance=self._config.distance)

    def _filter(self, filter, limit=10, **kwargs):
        kwargs = self._update_kwargs_from_config('consistency_level', **kwargs)
        with self.loaded_collection():
            results = self._collection.query(
                expr=filter,
                limit=limit,
                output_fields=['serialized'],
                **kwargs,
            )
        return self._docs_from_query_response(results)[:limit]
